#pragma once

#include "volumeshop.h"

#include "RendererPlugin.h"
#include "PluginInstance.h"
#include "Renderer.h"
#include "Exception.h"
#include "Shader.h"
#include "ResourceTexture2D.h"

class RendererTutorial : public Renderer, public PluginInstance<RendererPlugin>
{

public:

	RendererTutorial(Plugin & pluPlugin) : PluginInstance(pluPlugin), m_shaShader(pluPlugin.GetPluginDirectory()+"plugin_renderer_tutorial.glsl")
	{
		m_modUpdateObserver.connect(this,&RendererTutorial::triggerUpdate);

		m_shaShader.addObserver(&m_modUpdateObserver);
		GetPlugin().GetProperty("Projection Transformation").require(Variant::TypeMatrix()).addObserver(&m_modUpdateObserver);
		GetPlugin().GetProperty("Viewing Transformation").require(Variant::TypeMatrix()).addObserver(&m_modUpdateObserver);
		GetPlugin().GetProperty("Color").require(Variant::TypeColor(Color(1.0f,1.0f,1.0f))).addObserver(&m_modUpdateObserver);
		GetPlugin().GetProperty("Texture Map").require(Variant::TypeHandle()).addObserver(&m_modUpdateObserver);
		GetPlugin().GetProperty("Displacement Map").require(Variant::TypeHandle()).addObserver(&m_modUpdateObserver);
		GetPlugin().GetProperty("Displacement").require(Variant::TypeFloat(0.0f,-32.0f,32.0f)).addObserver(&m_modUpdateObserver);

		m_pSphere = gluNewQuadric();

		if (!m_pSphere)
			throw Exception("Error generating quadric!");

		gluQuadricOrientation(m_pSphere,GLU_OUTSIDE);
		gluQuadricNormals(m_pSphere,GLU_SMOOTH);
		gluQuadricTexture(m_pSphere,GL_TRUE);

	};
 
	virtual ~RendererTutorial()
	{
		if (m_pSphere)
		{
			gluDeleteQuadric(m_pSphere);
			m_pSphere = NULL;
		}	
	};

	virtual void idle()
	{
	};

	virtual void display(Canvas & canCanvas)
	{
		const Matrix matProjection = GetPlugin().GetProperty("Projection Transformation");
		const Matrix matViewing = GetPlugin().GetProperty("Viewing Transformation");
		const Color colColor = GetPlugin().GetProperty("Color");
		
		Handle hanTextureMap = GetPlugin().GetProperty("Texture Map");
		ImageResource *pTextureMap = hanTextureMap.GetResource<ImageResource>();
		
		Handle hanDisplacementMap = GetPlugin().GetProperty("Displacement Map");
		ImageResource *pDisplacementMap = hanDisplacementMap.GetResource<ImageResource>();

		const float fDisplacement = GetPlugin().GetProperty("Displacement");

		canCanvas.bind();

		glMatrixMode(GL_PROJECTION);
		glLoadMatrixf(matProjection.Get());

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_LESS);

		const float vfSpecularReflectance[] = { 0.6f, 0.6f, 0.6f, 1.0f };
		glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR, vfSpecularReflectance);
		
		const float fShininess = 32.0f;
		glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,fShininess);
	
		const float vfLightPosition[] = { 0.0f, 0.0f, 1.0f, 0.0f };
		glLightfv(GL_LIGHT0, GL_POSITION, vfLightPosition);

		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glEnable(GL_NORMALIZE);

		glEnable(GL_COLOR_MATERIAL);
		glColorMaterial(GL_FRONT_AND_BACK,GL_DIFFUSE);
		glColor4ubv(colColor.Get());

		glLoadMatrixf(matViewing.Get());

		m_shaShader.bind();

		if (pTextureMap)
		{
			glActiveTexture(GL_TEXTURE0);
			ResourceTexture2D::Get<ImageTexture2D>(pTextureMap).bind();

			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

			glUniform1i(m_shaShader.GetUniformLocation("samTexture"),0);
		}

		if (pDisplacementMap)
		{
			glActiveTexture(GL_TEXTURE1);
			ResourceTexture2D::Get<ImageTexture2D>(pDisplacementMap).bind();

			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

			glUniform1i(m_shaShader.GetUniformLocation("samDisplacement"),1);
		}

		glUniform1f(m_shaShader.GetUniformLocation("fDisplacement"),fDisplacement);

		gluSphere(m_pSphere,128.0,256,256);

		if (pDisplacementMap)
		{
			glActiveTexture(GL_TEXTURE1);
			ResourceTexture2D::Get<ImageTexture2D>(pDisplacementMap).release();
		}

		if (pTextureMap)
		{
			glActiveTexture(GL_TEXTURE0);
			ResourceTexture2D::Get<ImageTexture2D>(pTextureMap).release();
		}

		m_shaShader.release();

		canCanvas.release();
	};

	virtual void reshape(const unsigned int uWidth, const unsigned int uHeight)
	{
		GetPlugin().update();
	};

	virtual void triggerUpdate(const Observable & obsObservable, const Observable::Event & eveEvent)
	{
		GetPlugin().update();
	};

private:

	GLUquadric *m_pSphere;
	Shader m_shaShader;
	ModifiedObserver m_modUpdateObserver;
};

DECLARE_RENDERERPLUGIN(RendererTutorial)